############################################################################
#                                                                          #
# Imports & Constants                                                      #
#                                                                          #
############################################################################

import gpudb
import csv
import sys
import json

HOST = "127.0.0.1"
PORT = "9191"
CSV = "nyc_neighborhood.csv"

OPTION_NO_ERROR = {"no_error_if_not_exists": "true"}

TABLE_NYC_N = "nyc_neighborhood"
TABLE_TAXI = "nyctaxi"
TABLE_TAXI_EW = TABLE_TAXI + "_edges_weights_ids"
TABLE_TAXI_N = TABLE_TAXI + "_nodes"
TABLE_TAXI_N_S = TABLE_TAXI_N + "_sharded"

JOIN_TAXI = "taxi_tables_joined"
JOIN_PR_RESULTS = "page_rank_results_joined"

GRAPH_T = TABLE_TAXI + "_graph_id"
GRAPH_T_PRSOLVED = GRAPH_T + "_page_rank_solved"
GRAPH_T_PRSOLVED_S = GRAPH_T_PRSOLVED + "_sharded"
TABLE_GRAPH_T = GRAPH_T + "_table"

############################################################################
#                                                                          #
# Table Setup Function                                                     #
#                                                                          #
############################################################################

def table_setup():
    """ Setup necessary source tables for graph solver examples. """

    print("")
    print("===========")
    print("TABLE SETUP")
    print("===========")
    print("")

    # Clear any related tables in case they already exist
    kinetica.clear_table(table_name=JOIN_TAXI, options=OPTION_NO_ERROR)
    kinetica.clear_table(table_name=TABLE_NYC_N, options=OPTION_NO_ERROR)
    kinetica.clear_table(table_name=TABLE_TAXI_EW, options=OPTION_NO_ERROR)
    kinetica.clear_table(table_name=TABLE_TAXI_N, options=OPTION_NO_ERROR)

    # Define the NYC neighborhood table type
    nycn_columns = [
        ["gid", "int"],
        ["geom", "string", "wkt"],
        ["CTLabel", "string", "char16"],
        ["BoroCode", "string", "char16"],
        ["BoroName", "string", "char16"],
        ["CT2010", "string", "char16"],
        ["BoroCT2010", "string", "char16"],
        ["CDEligibil", "string", "char16"],
        ["NTACode", "string", "char16"],
        ["NTAName", "string", "char64"],
        ["PUMA", "string", "char16"],
        ["Shape_Leng", "double"],
        ["Shape_Area", "double"]
    ]

    # Create the NYC neighborhood table
    try:
        table_nycn_obj = gpudb.GPUdbTable(
            _type=nycn_columns,
            name=TABLE_NYC_N,
            db=kinetica,
            options={"is_replicated": "true"}
        )
        print("{} table object successfully created.".format(TABLE_NYC_N))
    except gpudb.GPUdbException as e:
        print(
            "{} table object creation failure: {}".format(TABLE_NYC_N, str(e))
        )

    # Increase the field size limit because CSV contains large field values
    csv.field_size_limit(sys.maxsize)

    # Insert records from a local CSV file into the NYC neighborhood table
    print("Creating records from the " + CSV + " file.")
    nycn_data = csv.reader(open(CSV))
    nycn_data.next()
    nycn_records = []
    for record in nycn_data:
        record_data = [int(record[0]), record[1], record[2], record[3],
                       record[4], record[5], record[6], record[7], record[8],
                       record[9], record[10], float(record[11]),
                       float(record[12])]
        nycn_records.append(record_data)
    print(
        "Inserting records into {} table. This may take a "
        "while.".format(TABLE_NYC_N)
    )
    table_nycn_obj.insert_records(nycn_records)
    print(
        "Number of records inserted into the {} "
        "table: {}".format(TABLE_NYC_N, table_nycn_obj.size())
    )
    print("")

    # Check to see if 'nyctaxi' table exists
    print("Checking to see if table {} exists.".format(TABLE_TAXI))
    nyctaxi_status = kinetica.has_table(
        table_name=TABLE_TAXI
    )['table_exists']
    if nyctaxi_status:
        print("Table {} exists.".format(TABLE_TAXI))
    else:
        print(
            "Table {} does not exist. Please ingest the table (in "
            "GAdmin, Cluster > Demo > NYC Taxi > Load Sample Data) and try "
            "again.".format(TABLE_TAXI))
        sys.exit(1)
    print("")

    # Join the TABLE_TAXI table to the TABLE_NYC_N table using STXY_CONTAINS
    # to filter out data that could skew the graph
    print(
        "Joining {} to {} to filter out data that could skew the taxi "
        "graphs.".format(TABLE_TAXI, JOIN_TAXI)
    )
    join_taxi_tables_response = kinetica.create_join_table(
        join_table_name=JOIN_TAXI,
        table_names=[TABLE_TAXI + " as t", TABLE_NYC_N + " as n"],
        column_names=[
            "CONCAT(CHAR32(pickup_longitude), CHAR32(pickup_latitude)) as pickup_name",
            "t.pickup_longitude", 
            "t.pickup_latitude",
            "HASH(t.pickup_longitude + t.pickup_latitude) as pickup_id",
            "CONCAT(CHAR32(dropoff_longitude), CHAR32(dropoff_latitude)) as dropoff_name",
            "t.dropoff_longitude",
            "t.dropoff_latitude",
            "HASH(t.dropoff_longitude + t.dropoff_latitude) as dropoff_id",
            "t.total_amount"
        ],
        expressions=[
            "(STXY_CONTAINS(n.geom, t.pickup_longitude, t.pickup_latitude)) AND"
            "(STXY_CONTAINS(n.geom, t.dropoff_longitude, t.dropoff_latitude)) "
        ]
    )["status_info"]["status"]
    print("{} view created: {}".format(JOIN_TAXI, join_taxi_tables_response))
    print("")

    # Union the JOIN_TAXI view to itself to collapse pickup & dropoff
    # locations into a unified set of endpoint locations to serve as node IDs
    print(
        "Unioning {} to itself contain the graph's nodes.".format(
            JOIN_TAXI
        )
    )
    sql = "CREATE TABLE " + TABLE_TAXI_N + " AS " \
          "SELECT " \
                "pickup_id as id, " \
                "pickup_longitude as lon, " \
                "pickup_latitude as lat " \
          "FROM " + JOIN_TAXI + " " \
          "UNION " \
          "SELECT " \
                "dropoff_id, " \
                "dropoff_longitude, " \
                "dropoff_latitude " \
          "FROM " + JOIN_TAXI
    nodes_response = kinetica.execute_sql(
        statement=sql,
        offset=0,
        limit=gpudb.GPUdb.END_OF_SET,
        encoding="json",
        options={}
    )["status_info"]["status"]
    print("{} union created: {}".format(TABLE_TAXI_N, nodes_response))
    print("")

    # Create a projection to contain the graph edges (based on NODE_ID)
    print(
        "Creating a projection from {} to contain the graph's edges.".format(
            JOIN_TAXI
        )
    )
    edges_id_response = kinetica.create_projection(
        table_name=JOIN_TAXI,
        projection_name=TABLE_TAXI_EW,
        column_names=["pickup_id", "dropoff_id"]
    )["status_info"]["status"]
    print(
        "{} projection created: {}".format(
            TABLE_TAXI_EW, edges_id_response
        )
    )

# end table_setup()

############################################################################
#                                                                          #
# Graph Setup Function                                                     #
#                                                                          #
############################################################################

def graph_setup():
    """ Setup graphs to be used in solver examples. """

    print("")
    print("===========")
    print("GRAPH SETUP")
    print("===========")
    print("")

    # Create a graph from TABLE_TAXI_N and TABLE_TAXI_EW using IDs
    print("Creating {}".format(GRAPH_T))
    create_t_graph_response = kinetica.create_graph(
        graph_name=GRAPH_T,
        directed_graph=False,
        nodes=[
            TABLE_TAXI_N + ".id AS NODE_ID"
        ],
        edges=[
            TABLE_TAXI_EW + ".pickup_id AS EDGE_NODE1_ID",
            TABLE_TAXI_EW + ".dropoff_id AS EDGE_NODE2_ID"
        ],
        weights=[],
        restrictions=[],
        options={
            "recreate": "true"
        }
    )
    print("{} graph created: {}".format(
        GRAPH_T, create_t_graph_response["status_info"]["status"])
    )
    print("Number of nodes: {}".format(create_t_graph_response["num_nodes"]))
    print("Number of edges: {}".format(create_t_graph_response["num_edges"]))

# end graph_setup()

def page_rank_example():
    """Demonstrate creating and solving a graph using the PAGE_RANK method.
    """

    print("")
    print("========================")
    print("PAGE RANK SOLVER EXAMPLE")
    print("========================")
    print("")

    # Clear any related tables in case they already exist
    kinetica.clear_table(
        table_name=GRAPH_T_PRSOLVED, options=OPTION_NO_ERROR
    )
    kinetica.clear_table(table_name=TABLE_TAXI_N_S, options=OPTION_NO_ERROR)
    kinetica.clear_table(
        table_name=GRAPH_T_PRSOLVED_S, options=OPTION_NO_ERROR
    )
    kinetica.clear_table(
        table_name=JOIN_PR_RESULTS, options=OPTION_NO_ERROR
    )

    # Solve the graph
    print("Solving {} using PAGE_RANK solver type.".format(GRAPH_T))
    solve_pr_graph_response = kinetica.solve_graph(
        graph_name=GRAPH_T,
        solver_type="PAGE_RANK",
        source_nodes=["129341667930495514"],
        destination_nodes=[],
        solution_table=GRAPH_T_PRSOLVED,
        options={}
    )["status_info"]["status"]
    print("{} graph solved: {}".format(
        GRAPH_T, solve_pr_graph_response)
    )
    print("")

    # Shard the TABLE_TAXI_N table so it can be joined to the Page Rank results
    # table
    nodes_sharded_response = kinetica.create_projection(
        table_name=TABLE_TAXI_N,
        projection_name=TABLE_TAXI_N_S,
        column_names=["id", "lon", "lat"],
        options={"shard_key": "id"}
    )["status_info"]["status"]
    print(
        "{} projection created: {}".format(
            TABLE_TAXI_N_S, nodes_sharded_response
        )
    )

    # Shard the Page Rank results table so it can be joined to the TABLE_TAXI_N
    # table
    graph_sharded_response = kinetica.create_projection(
        table_name=GRAPH_T_PRSOLVED,
        projection_name=GRAPH_T_PRSOLVED_S,
        column_names=["SOLVERS_NODE_ID", "SOLVERS_NODE_COSTS"],
        options={"shard_key": "SOLVERS_NODE_ID"}
    )["status_info"]["status"]
    print(
        "{} projection created: {}".format(
            GRAPH_T_PRSOLVED_S, graph_sharded_response
        )
    )

    # Join the TABLE_TAXI_N_S and GRAPH_T_PRSOLVED_S tables to pair the page
    # rank results IDs and costs with the longitude/latitude pair of the node
    join_pr_response = kinetica.create_join_table(
        join_table_name=JOIN_PR_RESULTS,
        table_names=[
            TABLE_TAXI_N_S + " as n",
            GRAPH_T_PRSOLVED_S + " as s"
        ],
        column_names=[
            "n.lon", 
            "n.lat", 
            "s.SOLVERS_NODE_ID", 
            "s.SOLVERS_NODE_COSTS"
        ],
        expressions=["n.id = s.SOLVERS_NODE_ID"],
        options={}
    )["status_info"]["status"]
    print("{} join view created: {}".format(JOIN_PR_RESULTS, join_pr_response))
    print("")

    # Retrieve the top 10 nodes in terms of most visited point
    print("Top 10 nodes sorted by cost (highest to lowest):")
    print(
        "{:>9} {:>8} {:>20} {:>15}".format("Longitude", "Latitude", "ID",
                                           "Cost")
    )
    print("{:=>9} {:=>8} {:=>20} {:=>15}".format("", "", "", ""))
    get_pr_response = kinetica.get_records(
        table_name=JOIN_PR_RESULTS,
        offset=0,
        limit=10,
        encoding="json",
        options={"sort_by": "SOLVERS_NODE_COSTS", "sort_order": "descending"}
    )["records_json"]
    for record in get_pr_response:
        print("{lon:>9} {lat:>8} {SOLVERS_NODE_ID:>20} "
              "{SOLVERS_NODE_COSTS:<15}".format(**json.loads(record))
              )

# end page_rank_example()



if __name__ == '__main__':
    # Establish connection with a locally-running instance of Kinetica
    kinetica = gpudb.GPUdb(encoding='BINARY', host=HOST, port=PORT)

    # Execute defined functions
    table_setup()
    graph_setup()
    page_rank_example()
